from requests import get
import bs4

class PageNotFoundException(Exception):
    """Raised when a page response is not 200"""
    
class TooFewLinksException(Exception):
    """Raised when the function requests more then the number of links available on a page"""
    
    def __init__(self, link_list: list[str]) -> None:
        self.link_list = link_list
        super().__init__()


def get_page_html(url: str) -> str:
    """Gets the html content of a page

    Args:
        url (str): url of page to get

    Returns:
        str: html of page

    Raises:
        PageNotFoundException
    """
    page_response = get(url)
    if page_response.status_code != 200:
        raise PageNotFoundException

    return page_response.content


def get_links(html: str, url_range: int) -> list[str]:
    """Gets first <range> links in a wikipedia html document that are not in the note section or in parenthesis

    Args:
        html (str): html of wikipedia page
        range (int): number of links to get

    Returns:
        list[str]: list of links in document
        
    Raises:
        TooFewLinksException
    """
    soup = bs4.BeautifulSoup(html, "html.parser")

    main_content: bs4.Tag = soup.find(id="mw-content-text").find("div")

    for decompose_type in ("div", "style", "sup", "span", "table", "figure"):
        for item in main_content.find_all(decompose_type):
            item.decompose()

    for section in main_content:
        paren_level = 0
        for child_section in section:
            child_text = ""
            
            match type(child_section):
                case bs4.element.NavigableString:
                    child_text = child_section
                case bs4.element.Tag:
                    child_text = child_section.text
                    
            if "(" in child_text and ")" in child_text:
                continue
            elif "(" in child_text:
                paren_level += 1
            elif ")" in child_text:
                paren_level -= 1
            if paren_level > 0 and isinstance(child_section, bs4.element.Tag):
                blank = soup.new_tag("b")
                child_section.replace_with(
                    blank
                )  # prevents iterator from skipping a tag due to decomposition
    links = [
        "https://en.wikipedia.org" + l
        for l in [t.get("href") for t in main_content.find_all("a", limit=url_range)]
        if "en.wiktionary.org" not in l
    ]
    
    if len(links) < url_range - 1:
        raise TooFewLinksException(links)

    return links



from rich import print
from typer import Exit
import graphviz

from .lib import PageNotFoundException, TooFewLinksException, get_links, get_page_html
from . import app

@app.command()
def psychology(url: str, max_length: int = 15):
    """
    Scrapes the leading hyperlink of Wikipedia pages recursively until the page "Psychology" is reached.
    """
    current_url = url
    current_url_split = current_url.split("/")
    print(f"Starting page: {'/'.join(current_url_split[:-1])}[bold red]/{current_url_split[-1]}[/bold red]  [link={current_url}]|Link|[/link]")
    
    flag_reached = False
    print_color = "yellow"
    counter = 0
    while not flag_reached and counter != max_length:
        try:
            current_url = get_links(get_page_html(current_url), 1)[0]
        except PageNotFoundException:
            print("[red]Requested page not found![/red]")
        current_url_split = current_url.split("/")
        
        if current_url == "https://en.wikipedia.org/wiki/Psychology" or current_url == "https://en.wikipedia.org/wiki/Philosophy":
            flag_reached = True
            print_color = "green"

        print(f"-> {'/'.join(current_url_split[:-1])}[bold {print_color}]/{current_url_split[-1]}[/bold {print_color}]  [link={current_url}]|Link|[/link]")
        
        counter += 1
        
@app.command()
def graph_multiple(urls: list[str], max_length: int = 15, verbose: bool = False):
    """
    Given multiple wikipedia pages, scrapes and generated a directed graphviz source file showing the those articles' path to Psychology.
    For ease of use, direct the output of this command into a graphviz dot file.
    """
    graph = graphviz.Digraph()

    for url in urls:
        current_url = url
        current_url_name = current_url.split("/")[-1]
        graph.node(current_url.replace("https://", ""), current_url_name)
        
        flag_reached = False
        counter = 0
        while not flag_reached and counter != max_length:
            old_url = current_url
            
            try:
                current_url = get_links(get_page_html(current_url), 1)[0]
            except PageNotFoundException:
                print("[red]A requested page was not found![/red]")
            
            current_url_name = current_url.split("/")[-1]
            current_url_right = current_url.replace("https://", "")
            graph.node(current_url_right, current_url_name)
            graph.edge(old_url.replace("https://", ""), current_url_right)
            
            if verbose:
                print(f"[yellow]{old_url}[/yellow] -> [yellow]{current_url}[/yellow]")
            
            if current_url == "https://en.wikipedia.org/wiki/Psychology" or current_url == "https://en.wikipedia.org/wiki/Philosophy":
                if verbose:
                    print("[green]Source reached Psychology/Philosophy[/green]")
                break
            counter += 1
            
    print("Graphviz source:")
    print(graph.source)
    
@app.command()
def scrape(url: str, url_range: int = 10):
    """
    Scrape a Wikipedia page for its hyperlinks. Ignores the topmost note section and links in parenthetical sentences.
    """
    try:
        page_html = get_page_html(url)
    except PageNotFoundException:
        print("[red]Requested page not found![/red]")
        Exit(1)
        
    try:
        link_list = get_links(page_html, url_range)
    except TooFewLinksException as err:
        print("[red]Too many links requested. Printing every link on page.[/red]")
        link_list = err.link_list
        
    print("\n".join(link_list))

if __name__ == "__main__":
    app()


import pytest
from wikipedia_scraper.lib import PageNotFoundException, TooFewLinksException, get_page_html, get_links

@pytest.fixture
def page_html():
    with open("tests/wikipedia_source.html") as file:
        return file.read()

def test_request():
    html = get_page_html("https://en.wikipedia.org/wiki/Organism")
    assert b"Organism" in html
    
def test_request_fail():
    with pytest.raises(PageNotFoundException):
        get_page_html("https://en.wikipedia.org/wiki/Goofy_Ani_O")

def test_get_links(page_html):
    links_c = get_links(page_html, 4)
    assert links_c == [
        "https://en.wikipedia.org/wiki/Country_music",
        "https://en.wikipedia.org/wiki/Pop_music",
        "https://en.wikipedia.org/wiki/Nashville_sound",
        "https://en.wikipedia.org/wiki/Countrypolitan"
    ]

def test_link_few_fail(page_html):
    with pytest.raises(TooFewLinksException):
        get_links(page_html, 100000)